@@ -1,12 +1,42 @@
 """
-Control complex python object through strings.
-Wrap functions so that the complex arguments can be specified through a string key
+CRUDE stands for CRUD-Execution.
+It is a method to solve the problem of dealing with complex python objects in an
+environment that doesn't natively support these.
+
+The method's trick is to allow the complex object's that we "crudified" to be controlled
+via a string key that references the complex object, via a "store" which maps
+these string keys to the actual physical object.
+This store could be a python dictionary (so in RAM) or any persisting storage system
+(files, DB) that is given a `typing.Mapping` interface
+(see https://i2mint.github.io/dol/ or https://i2mint.github.io/py2store for
+tools to do so).
+
+Take, for instance, a GUI that allows a user to compute some descriptive statistics
+of the columns of a table.
+The inputs are a table, and one of the following statistics function:
+``statistics.mean``, ``statistics.median``, or ``statistics.stdev``.
+
+Python functions are not a type natively handled by GUI, so what can we do?
+We can stick a layer between our ``compute_stats(stats_func, table)`` function
+and our GUI, endowed with a
+``{"mean": statistics.mean``, "median": statistics.median, "std": statistics.stdev}``
+mapping. We expose the string keys to the GUI, and map them to the functions before
+calling ``compute_stats``.
+
+In the case of the ``table``, we'd probably add a means for the GUI user to upload
+tables (say from ``.csv`` or ``.xlsx`` files), storing them under a name of their
+choice, then pointing to said table via the name, when they want to execute a
+``compute_stats(stats_func, table)``.
+
+These are examples of what we call "crudifying" variables or functions.
+
+Here we therefore offer tools to do this sort of thing;
+wrap functions so that the complex arguments can be specified through a string key
 that points to the actual python object (which is stored in a session's memory or
 persisted in some fashion).
 """
 
-from typing import Any, Mapping, Optional, Callable, Union, Iterable
-from functools import partial
+from typing import Any, Mapping, Optional, Callable, Union, Iterable, Iterator
 from inspect import Parameter
 import os
 
@@ -87,9 +117,9 @@ def store_on_output(
     store=None,
     save_name_param='save_name',
     add_store_to_func_attr='output_store',
-    empty_name_callback=None,
-    auto_namer=None,
-    output_trans=None,
+    empty_name_callback: Callable[[], Any] = None,
+    auto_namer: Callable[..., str] = None,
+    output_trans: Callable[..., Any] = None,
 ):
     """Wrap func so it will have an extra save_name_param that can be used to
     indicate whether to save the output of the function call to that key, in
@@ -198,6 +228,9 @@ def store_on_output(
         arguments = sig.kwargs_from_args_and_kwargs(args, kwargs, apply_defaults=True)
         save_name = arguments.pop(save_name_param)
         if not save_name and empty_name_callback:
+            assert callable(
+                empty_name_callback
+            ), f'empty_name_callback must be callable: {empty_name_callback}'
             empty_name_callback()
         args, kwargs = sig.args_and_kwargs_from_kwargs(arguments)
         output = func(*args, **kwargs)
@@ -222,11 +255,15 @@ def store_on_output(
 def prepare_for_crude_dispatch(
     func: Callable = None,
     *,
-    param_to_mall_map: Optional[Iterable] = None,
+    param_to_mall_map: Optional[Union[dict, Iterable]] = None,
     mall: Optional[Mall] = None,
-    output_store: Optional[Union[Mapping, str]] = None,
-    save_name_param: str = 'save_name',
     include_stores_attribute: bool = False,
+    output_store: Optional[Union[Mapping, str]] = None,
+    # the arguments below only apply if output_store is given
+    save_name_param: str = 'save_name',
+    empty_name_callback: Callable[[], Any] = None,
+    auto_namer: Callable[..., str] = None,
+    output_trans: Callable[..., Any] = None,
 ):
     """
     Wrap func into something that is ready for CRUDE dispatch.
@@ -240,15 +277,31 @@ def prepare_for_crude_dispatch(
         (e.g. dict) that should be used for said param.
         If a non-Mapping iterable is given, will take {name: name...} identity mapping
         for names in that iterable.
-    :mall mall: A store of stores. A Mapping whose keys are what the values of
+    :param mall: A store of stores. A Mapping whose keys are what the values of
         ``param_to_mall_map`` point to and whose values are mapping interfaces (called
          "stores" of a storage backend (local or remote, persisted or in-memory).
+    :param include_stores_attribute: bool, whether to add an attribute to the function
+        containing the ``output_store``
     :param output_store: a store used to record the output of the function
+
+    If (and only if) ``output_store`` is given, we can also use the following
+    arguments to control the output "crudification" further.
+
     :param save_name_param: str, the argument name that should be used in the returned
         functions to get the the key of ``output_store`` under which the output will be
         saved.
-    :param include_stores_attribute: bool, whether to add an attribute to the function
-        containing the ``output_store``
+    :param empty_name_callback: If not None, will be called when the user (of the
+        wrapped function ``prepare_for_crude_dispatch`` will output) doesn't
+        specify a save name. Intended use; raising errors (or other kinds of behavior)
+        when we want to force the user to enter name.
+    :param auto_namer: If not None, should be a keyword-only ``(*, arguments, output)``
+        callback that will be called (forgivingly). This callback will
+        be called when the user doesn't specify a name (or an empty name). It's output
+        will be used as the ``save_name``. Intended use is to produce a function where
+        that has auto-naming capabilities.
+    :param output_trans: If not None, should be a keyword-only
+        ``(*, save_name, output, arguments)`` function that will be called, returning
+        it's result instead of the ``output``.
 
     :return: A function that outputs the same thing as ``func``, but (1) with some
         parameters being changed so that on can specify some arguments
@@ -329,6 +382,35 @@ def prepare_for_crude_dispatch(
     >>> output_store
     {'save_here': 31}
 
+    The example below shows how one can
+
+    - change the default ``save_name_param``
+
+    - use an ``auto_namer``
+
+    - use ``output_trans`` to return both save name and output
+
+    >>> def bar(a, b: int = 2):
+    ...     return a + b
+    >>> my_store = {'all': 'mine'}
+    >>> def return_key_and_val(*, save_name, output):
+    ...     return save_name, output
+    >>> wbar = prepare_for_crude_dispatch(
+    ...     bar,
+    ...     output_store=my_store,
+    ...     save_name_param='save_as',
+    ...     auto_namer=lambda *, arguments: '_'.join(map(str, arguments.values())),
+    ...     output_trans=return_key_and_val
+    ... )
+
+    >>> wbar(2, 3, save_as='test')  # output will be the (save_as, output) pair
+    ('test', 5)
+    >>> my_store  # but only the output is saved (under save_as key)
+    {'all': 'mine', 'test': 5}
+    >>> wbar(7)  # and if you don't specify a save_as key, one is made for you!
+    ('7_2', 9)
+    >>> my_store
+    {'all': 'mine', 'test': 5, '7_2': 9}
     """
     ingress = None
 
@@ -338,6 +420,7 @@ def prepare_for_crude_dispatch(
 
         sig = Sig(func)
 
+        # get an {argname: store, ...} dict from param_to_mall_map:
         store_for_param = _mk_store_for_param(sig, param_to_mall_map, mall) or dict()
 
         def kwargs_trans(outer_kw):
@@ -376,7 +459,7 @@ def prepare_for_crude_dispatch(
             inner_sig=sig,
             kwargs_trans=kwargs_trans,
             outer_sig=(
-                sig.ch_annotations(**{name: str for name in param_to_mall_map})
+                sig.ch_annotations(**{name: str for name in store_for_param})
                 # + [save_name_param]
             ),
         )
@@ -402,16 +485,15 @@ def prepare_for_crude_dispatch(
             )
 
         wrapped_f = store_on_output(
-            wrapped_f, store=output_store, save_name_param=save_name_param,
+            wrapped_f,
+            store=output_store,
+            save_name_param=save_name_param,
+            add_store_to_func_attr='output_store' if include_stores_attribute else None,
+            empty_name_callback=empty_name_callback,
+            auto_namer=auto_namer,
+            output_trans=output_trans,
         )
-        wrapped_f.__name__ = wrapped_f.__name__ + '_w_output_storing'
-
-        if include_stores_attribute:
-            wrapped_f.output_store = output_store
-
-        # def egress(func_output):
-        #     store_for_param[output_store_name] = func_output
-        #     return func_output
+        # wrapped_f.__name__ = wrapped_f.__name__ + "_w_output_storing"  # to remove
 
     return wrapped_f
 
